"""
selenium.
"""
def selenium_support(spider_cls):
    """spider 装饰器

    添加 selenium 相关设置
    使得在 spider 中可使用 SeleniumRequest

    """
    from scrapy_ajax_utils.utils import add_settings_to_spider

    settings = {
        'DOWNLOADER_MIDDLEWARES': {'scrapy_ajax_utils.selenium.SeleniumDownloadMiddleWare': 600}
    }
    return add_settings_to_spider(spider_cls, settings)


# 便于其它文件导入
from .driver import webdriver
from .request import SeleniumRequest
from .middleware import SeleniumDownloadMiddleWare

from selenium.webdriver.common.by import By

SeleniumBy = By


import logging

from scrapy import signals
from scrapy.http import HtmlResponse
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from scrapy_ajax_utils.utils import extract_domain_from_url
from scrapy_ajax_utils.selenium.driver import webdriver
from scrapy_ajax_utils.selenium.request import SeleniumRequest

logger = logging.getLogger(__name__)


class SeleniumDownloadMiddleWare(object):
    """For selenium.

    注意：
        缓存cookies需要浏览器User-Agent请求头版本与设置脚本(或Request)中的默认请求头保持一致
        否则某些网站可能会对此做验证 导致cookies无效

    """

    def __init__(self, settings):
        self.settings = settings
        self._driver = None
        self._cached_cookies = {}

    @property
    def driver(self):
        if self._driver is None:
            self._driver = self._get_driver()
        return self._driver

    def _get_driver(self):
        headless = self.settings.getbool('SELENIUM_HEADLESS', True)
        driver_name = self.settings.get('SELENIUM_DRIVER_NAME', 'chrome')
        executable_path = self.settings.get('SELENIUM_DRIVER_PATH')
        return webdriver(driver_name=driver_name,
                         headless=headless,
                         executable_path=executable_path)

    @classmethod
    def from_crawler(cls, crawler):
        dm = cls(crawler.settings)
        crawler.signals.connect(dm.closed, signal=signals.spider_closed)
        return dm

    def process_request(self, request, spider):
        if not isinstance(request, SeleniumRequest):
            return

        if request.cache_cookies:
            for domain in self._cached_cookies:
                if domain in request.url:
                    request.cookies = self._cached_cookies[domain]
                    return

        self.driver.get(request.url)

        # 检查请求是否携带Cookies
        if request.cookies:
            for k, v in request.cookies.items():
                self.driver.add_cookie({'name': k, 'value': v})
            self.driver.get(request.url)

        if request.wait_until:
            condition = EC.presence_of_element_located(request.wait_until)
            WebDriverWait(self.driver, request.wait_time).until(condition)

        # Execute javascript code and save the result to meta.
        if request.script:
            result = self.driver.execute_script(request.script)
            if result is not None:
                request.meta['js_result'] = result

        if request.handler:
            request.handler(self.driver, request, spider)

        for cookie_name, cookie_value in request.cookies.items():
            self.driver.add_cookie(
                {
                    'name': cookie_name,
                    'value': cookie_value
                }
            )
        request.cookies = self.driver.get_cookies()
        request.meta['driver'] = self.driver

        if request.cache_cookies:
            domain = extract_domain_from_url(request.url)
            self._cached_cookies[domain] = request.cookies
        else:
            body = str.encode(self.driver.page_source)
            return HtmlResponse(
                self.driver.current_url,
                body=body,
                encoding='utf-8',
                request=request
            )

    def closed(self):
        if self._driver is not None:
            self._driver.quit()
            logger.debug('Selenium closed')



from selenium import webdriver as _webdriver
from scrapy.utils.project import get_project_settings


def webdriver(driver_name='chrome', executable_path=None,
              headless=True, disable_image=True, user_agent=None, options=None):
    if user_agent is None:
        settings = get_project_settings()
        user_agent = settings['DEFAULT_REQUEST_HEADERS'].get('User-Agent')

    if driver_name == 'firefox':
        if executable_path is None:
            executable_path = 'geckodriver'
        if options is None:
            options = _webdriver.FirefoxOptions()
            options.headless = headless
            if disable_image:
                options.set_preference('permissions.default.image', 2)
            options.set_preference('general.useragent.override', user_agent)
        # service_log_path: nul -> close log file, only work for windows.
        return _webdriver.Firefox(options=options, executable_path=executable_path, service_log_path='nul')

    elif driver_name == 'chrome':
        if executable_path is None:
            executable_path = 'chromedriver'
        if options is None:
            options = _webdriver.ChromeOptions()
            options.headless = headless
            options.add_argument(f"--user-agent={user_agent}")
            options.add_argument('--disable-gpu')
            if disable_image:
                options.add_experimental_option('prefs', {'profile.default_content_setting_values': {'images': 2}})
            # 规避检测
            options.add_experimental_option('excludeSwitches', ['enable-automation', ])
        return _webdriver.Chrome(options=options, executable_path=executable_path)

    else:
        raise ValueError(f'Not support driver name: {driver_name}')



from scrapy_splash import (
    SplashCookiesMiddleware as _SplashCookiesMiddleware,
    SplashRequest as _SplashRequest,
    SplashJsonResponse
)

from scrapy_ajax_utils.utils import add_settings_to_spider

SPLASH_SETTINGS = {
    'SPLASH_URL': 'http://127.0.0.1:8050/',
    'DOWNLOADER_MIDDLEWARES': {
        'scrapy_ajax_utils.splash.SplashCookiesMiddleware': 700,
        'scrapy_splash.SplashMiddleware': 723,
        'scrapy.downloadermiddlewares.httpcompression.HttpCompressionMiddleware': 810,
    },
    'SPIDER_MIDDLEWARES': {
        'scrapy_splash.SplashDeduplicateArgsMiddleware': 100,
    },
    'DUPEFILTER_CLASS': 'scrapy_splash.SplashAwareDupeFilter'
}


class SplashRequest(_SplashRequest):

    def __init__(self, url, *args, lua_script=None, js_script=None, keep_cookies=True, wait=2,
                 **kwargs):
        kwargs.setdefault('dont_send_headers', False)
        _args = kwargs.setdefault('args', {})
        _args['wait'] = wait
        is_exec = kwargs.setdefault('endpoint', 'execute')
        if is_exec == 'execute':
            if lua_script:
                _args['lua_source'] = lua_script
            else:
                _args['lua_source'] = render_lua_script(js_script, keep_cookies)
        super().__init__(url, *args, **kwargs)


class SplashCookiesMiddleware(_SplashCookiesMiddleware):
    """
    传递 cookies 给 request 交由 scrapy 处理
    """

    def process_response(self, request, response, spider):
        if not isinstance(response, SplashJsonResponse):
            return response

        if 'cookies' in response.data:
            request.cookies = response.data['cookies']

        return super().process_response(request, response, spider)


def splash_support(spider_cls):
    """spider 装饰器

    为 spider 添加 splash 相关设置

    """
    return add_settings_to_spider(spider_cls, SPLASH_SETTINGS)


def render_lua_script(js_script=None, keep_cookies=True):
    """生成lua脚本
    js脚本不支持传参，如果需要传参可参考下面的重写lua脚本
    如果有执行结果，返回字段为 js_result
    渲染的结果可能会有多余的空行，但不影响使用

    如果参数均为默认值，则完整脚本如下：
    `
        function main(splash, args)
            splash:init_cookies(splash.args.cookies)
            splash:go({
                splash.args.url,
                headers=splash.args.headers,
                http_method=splash.args.http_method,
                body=splash.args.body,
                })
            assert(splash:wait(args.wait))
            return {
                cookies = splash:get_cookies(),
                html = splash:html(),
            }
            end
    `
    :param js_script: js脚本
    :param keep_cookies: 是否传递cookies
    """
    if js_script:
        js_script = js_script.strip('\n')
        js_script = f"""
        local test = splash:jsfunc([[
        {js_script}
        ]])
        js_result = test()
        """
        js_result = 'js_result = js_result,'
    else:
        js_result = ''
    return f"""
    function main(splash, args)
        {'' if not keep_cookies else 'splash:init_cookies(splash.args.cookies)'}
        splash.images_enabled = false
        splash:go({{
            splash.args.url,
            headers=splash.args.headers,
            http_method=splash.args.http_method,
            body=splash.args.body,
            }})
        assert(splash:wait(args.wait))
        {js_script or ''}
        return {{
            {'' if not keep_cookies else 'cookies = splash:get_cookies(),'}
            html = splash:html(),
            {js_result}
        }}
    end
    """



import copy


def add_settings_to_spider(spider_cls, settings: dict):
    """为 spider 添加设置并返回 spider

    注意：
        如果设置中含有字典，且 spider.custom_settings 有同项设置
        则会更新 （即调用 dict 的 `update` 方法）
        否则会覆盖

    """
    s = copy.deepcopy(settings)
    cus_settings = spider_cls.custom_settings
    if cus_settings:
        for k, v in cus_settings.items():
            if isinstance(v, dict) and k in s:
                s[k].update(v)
            else:
                s[k] = v
    spider_cls.custom_settings = s
    return spider_cls


def extract_domain_from_url(url):
    d = url.split('://')[-1]
    d = d.split('/')[0]
    return d


import logging

import scrapy

logger = logging.getLogger(__name__)


class SeleniumRequest(scrapy.Request):
    """用于selenium的请求类

    需要开启selenium下载中间件

    """
    def __init__(self, url, callback=None,
                 wait_until=None, wait_time=10,
                 script=None, cache_cookies=True,
                 handler=None, **kwargs):

        # 等待条件，为元组
        # 仅支持： presence_of_element_located
        # 含两个元素，第一个为对应的类型（参考 selenium.webdriver.common.by.By）
        # 第二个元素为指定的条件
        # 如： `(By.ID, 'content')`
        self.wait_until = wait_until
        # `wait_until` 的等待时间
        self.wait_time = wait_time

        # js 脚本
        self.script = script

        # 是否缓存 cookies
        self.cache_cookies = cache_cookies

        # webdriver 处理函数
        # 该函数接收三个参数：
        #   1. `driver`，即 `webdriver.Chrome` 的实例
        #   2. `request`，当前 `SeleniumRequest` 的实例
        #   3. `spider`，当前 `Spider` 的实例
        #
        # 该函数将会在以下操作完成后调用：
        #   1. 成功请求目标地址
        #   2. 等待直到期望条件出现（如果有）
        #   3. 执行传入的 js 脚本（如果有）
        self.handler = handler

        super().__init__(url, callback, **kwargs)



# coding=utf-8

from setuptools import setup, find_packages

setup(
    name='scrapy_ajax_utils',
    version=0.012,
    description=(
        'ajax utils for scrapy.'
    ),
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    author='financial',
    author_email='1012593988@qq.com',
    maintainer='financial',
    maintainer_email='1012593988@qq.com',
    license='BSD License',
    packages=find_packages(),
    platforms=["all"],
    url='https://github.com/financialfly/scrapy_ajax_utils',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=[
            'scrapy>=1.7.3',
            'scrapy_splash>=0.7.2',
            'selenium>=3.141.0',
        ]
)